
package SFC::FleetSummary;

use strict;
use warnings qw{ all };

use POSIX;
use Clone qw{ clone };
use Getopt::Long;
use Carp;

use SFC::Ship;
use SFC::Fleet;
use SFC::Planet;
use SFC::Vector;

use SFC::ShipStats qw{
        speed
        is_ship
        stdsort
        drop_sort
        fuel_consumption
        cargo_space
        ore_cost
        crystal_cost
        hydrogen_cost
    };

# push @SFC::FleetSummary::ISA, 'Clone';

# sorted list of types
sub types {
    my $self = shift;

    return sort drop_sort $self->ship_types;
}


sub speed_of_slowest {
    my $self = shift;
    my $min = INT_MAX;

    for my $t ($self->ship_types) {
        next unless is_ship($t);
        my $s = speed($t, $self);
        $min = $s if $s < $min;
    }
    return $min;
}


# hydro cost for this fleet to travel distance
sub hydro_cost {
    my $self = shift;
    my ($distance, $frac) = @_;

    # determine "effective speed" - speed of slowest ship
    my $sp_eff = $self->speed_of_slowest();

    # sum over ship types
    my $cost = 0.0;
    for my $t ($self->ship_types) {
        next unless is_ship($t);

        $cost += $self->ship_count($t) * fuel_consumption($t, $self)
                * ($distance / 35000)
                * (1 + $frac * sqrt($sp_eff / speed($t, $self))) ** 2;
    }
    
    return ceil($cost);
}


sub hydro_cost_breakdown {
    my $self = shift;
    my ($distance, $frac) = @_;

    # determine "effective speed" - speed of slowest ship
    my $sp_eff = $self->speed_of_slowest();

    my %cost;
    for my $t ($self->ship_types) {
        next unless is_ship($t);

        $cost{$t} = $self->ship_count($t) * fuel_consumption($t, $self)
                * ($distance / 35000)
                * (1 + $frac * sqrt($sp_eff / speed($t, $self))) ** 2;
    }
    
    return \%cost;
}


sub summarize {
    my $self = shift;
    my ($fleet) = @_;
    my $summary = {};
    bless $summary, ref $self;

    for my $s (@{$fleet->ships()}) {
        my $t = $s->type();
        ++$summary->{$t}{count};
    }
}

sub ir_gain {
    my $self = shift;
    my ($amt, $ship) = @_;

    $self->{$ship->type()}{ir_gain} += $amt;
    $self->{ir_gain} += $amt;
}


sub ir_loss {
    my $self = shift;
    my ($amt, $ship) = @_;

    $self->{$ship->type()}{ir_loss} += $amt;
    $self->{ir_loss} += $amt;
}

sub overkill_given {
    my $self = shift;
    my ($amt, $ship) = @_;

    $self->{$ship->type()}{overkill_given} += $amt;
    $self->{overkill_given} += $amt;
}


sub overkill_taken {
    my $self = shift;
    my ($amt, $ship) = @_;

    $self->{$ship->type()}{overkill_taken} += $amt;
    $self->{overkill_taken} += $amt;
}


sub set_ships {
    my $self = shift;
    my ($type, $num) = @_;

    if (defined $num) {
        $self->{$type}{count} = $num;
    }
    elsif (ref($type) eq 'SFC::Vector') {
        my @x = @$type;
        for my $t ($self->types) {
            $self->{$t}{count} = shift @x;
        }
    }
}

sub ship_count {
    my $self = shift;
    my ($type) = @_;

    return $self->{$type}{count} if defined $type;

    my @x;
    for my $t ($self->types) {
        push @x, $self->{$t}{count};
    }
    return SFC::Vector->new(@x);
}

sub ship_types {
    grep /^[A-Z][a-z][a-z]$/, (keys %{$_[0]});
}

sub ship_lost {
    my $self = shift;
    my ($ship) = @_;
    my $type = $ship->type();
    
    --$self->{$type}{count};
    ++$self->{$type}{ships_lost};

    my $oc = ore_cost($type);
    my $cc = crystal_cost($type);
    my $hc = hydrogen_cost($type);
    my $tot = .6 * $oc + $cc + 1.5 * $hc;

    if ($ship->is_ship()) {
        $self->{ore_lost} += $oc;
        $self->{cry_lost} += $cc;
        $self->{gen_lost} += $hc;
        $self->{$type}{ore_loss} += $oc;
        $self->{$type}{cry_loss} += $cc;
        $self->{$type}{gen_loss} += $hc;
        $self->{$type}{loss} += $tot;
        $oc *= .3;
        $cc *= .3;
        $hc *= .3;
        $self->{debris} += $oc + $cc;
        $self->{deb_ore} += $oc;
        $self->{deb_cry} += $cc;
    }
    else {
        $oc *= .3;
        $cc *= .3;
        $hc *= .3;
        $self->{ore_lost} += $oc;
        $self->{cry_lost} += $cc;
        $self->{gen_lost} += $hc;
        $self->{$type}{ore_loss} += $oc;
        $self->{$type}{cry_loss} += $cc;
        $self->{$type}{gen_loss} += $hc;
        $self->{$type}{loss} += .3 * $tot;
    }
}

sub cargo {
    my $self = shift;

    my $cargo = 0;

    for my $t ($self->ship_types) {
        next unless is_ship($t);

        $cargo += $self->{$t}{count} * cargo_space($t);
    }

    return $cargo;
}


sub new {
    my $this = shift;
    my ($class, $self);
    if (ref($this)) {
        $class = ref($this);
        $self = clone $this;
    }
    else {
        $class = $this;
        $self = {
                # defaults
                id => 0,
                ore_lost => 0,
                cry_lost => 0,
                gen_lost => 0,
                debris => 0,
                deb_ore => 0,
                deb_cry => 0,
            };
    }
    bless $self, $class;
    while (@_) {
        my $arg = shift;
        if (ref($arg)) {
            while (my ($k, $v) = each %$arg) { $self->{$k} = $v; }
        }
        else {
            $self->{$arg} = shift;
        }
    }

    return $self;
}


sub get_attacker_options {
    my $self = shift;
   
    my $p = Getopt::Long::Parser->new(config => [ @_ ]);
    $p->getoptions($self,       # hash reference
            'file|att-file=s@',
            'weapon|att-weapon=i',
            'shield|att-shield=i',
            'armor|att-armor=i',
        ) or die 'Error:  parsing options';

    for my $f (@{$self->{file}}) {
        $self->read_file($f);
    }
}


sub get_defender_options {
    my $self = shift;
   
    my $p = Getopt::Long::Parser->new(config => [ @_ ]);
    $p->getoptions($self,       # hash reference
            'file|def-file=s@',
            'weapon|def-weapon=i',
            'shield|def-shield=i',
            'armor|def-armor=i',
        ) or die 'Error:  parsing options';

    for my $f (@{$self->{file}}) {
        $self->read_file($f);
    }
}


sub read_file {
    my $self = shift;
    my ($name, $is_target) = @_;
    my $f;

    $is_target = 1 unless defined $is_target;

    open $f, $name or croak "Can't open $name: $!"; 

    local $_;

    my $l  = qr/^(?:\* )?/;
    my $cl = qr/(?:[ _]+class)?/i;
    my $nr = qr/([\d,]+)\s*$/;

    while (<$f>) {

        my $q;
        if (/$nr/) {
            $q = $1;
            $q =~ s/,//g;
            $q += 0;
        }
        
        if ($is_target) {
            if (/${l}ore:?\s+$nr/i) {
                $self->{ore_plunder} = floor($q / 2);
                next;
            }
            if (/${l}crystal:?\s+$nr/i) {
                $self->{cry_plunder} = floor($q / 2);
                next;
            }
            if (/${l}hydrogen:?\s+$nr/i) {
                $self->{gen_plunder} = floor($q / 2);
                next;
            }
        }

        if (/${l}Shadow(?:[ _]+Probe)?:?\s*x?$nr/i) {
            $self->set_ships(Sha => $q);
            next;
        }
        if ($is_target &&
                $_ =~ /${l}Genesis(?:[ _]+Solar[ _]Satellite)?:?\s*x?$nr/i) {
            $self->set_ships(Gen => $q);
            next;
        }
        if (/${l}Hermes$cl(?:[ _]+Probe)?:?\s*x?$nr/i) {
            $self->set_ships(Hrm => $q);
            next;
        }
        if ($is_target &&
                $_ =~ /${l}Helios$cl(?:[ _]+Solar[ _]Satellite)?:?\s*x?$nr/i) {
            $self->set_ships(Hel => $q);
            next;
        }
        if (/${l}Artemis$cl(?:[ _]+Fighter)?:?\s*x?$nr/i) {
            $self->set_ships(Art => $q);
            next;
        }
        if (/${l}Atlas$cl(?:[ _]+Cargo)?:?\s*x?$nr/i) {
            $self->set_ships(Atl => $q);
            next;
        }
        if (/${l}Apollo$cl(?:[ _]+Fighter)?:?\s*x?$nr/i) {
            $self->set_ships(Apo => $q);
            next;
        }
        if (/${l}Zagreus$cl(?:[ _]+Recycler)?:?\s*x?$nr/i) {
            $self->set_ships(Zag => $q);
            next;
        }
        if (/${l}Charon$cl(?:[ _]+Transport)?:?\s*x?$nr/i) {
            $self->set_ships(Cha => $q);
            next;
        }
        if (/${l}Hercules$cl(?:[ _]+Cargo)?:?\s*x?$nr/i) {
            $self->set_ships(Hrc => $q);
            next;
        }
        if (/${l}Empusa$cl(?:[ _]+Fighter)?:?\s*x?$nr/i) {
            $self->set_ships(Emp => $q);
            next;
        }
        if (/${l}Dionysus$cl(?:[ _]+Recycler)?:?\s*x?$nr/i) {
            $self->set_ships(Dio => $q);
            next;
        }
        if (/${l}Curetes$cl(?:[ _]+Cruiser)?:?\s*x?$nr/i) {
            $self->set_ships(Cur => $q);
            next;
        }
        if (/${l}Poseidon$cl(?:[ _]+Cruiser)?:?\s*x?$nr/i) {
            $self->set_ships(Pos => $q);
            next;
        }
        if (/${l}Carmanor$cl(?:[ _]+Cargo)?:?\s*x?$nr/i) {
            $self->set_ships(Car => $q);
            next;
        }
        if (/${l}Pallas$cl(?:[ _]+Bomber)?:?\s*x?$nr/i) {
            $self->set_ships(Pal => $q);
            next;
        }
        if (/${l}Gaia$cl(?:[ _]+Colony[ _]+Ship)?:?\s*x?$nr/i) {
            $self->set_ships(Gai => $q);
            next;
        }
        if (/${l}Athena$cl(?:[ _]+Battleship)?:?\s*x?$nr/i) {
            $self->set_ships(Ath => $q);
            next;
        }
        if (/${l}Ares$cl(?:[ _]+Bomber)?:?\s*x?$nr/i) {
            $self->set_ships(Are => $q);
            next;
        }
        if (/${l}Hades$cl(?:[ _]+Battleship)?:?\s*x?$nr/i) {
            $self->set_ships(Had => $q);
            next;
        }
        if (/${l}Prometheus$cl(?:[ _]+Destroyer)?:?\s*x?$nr/i) {
            $self->set_ships(Pro => $q);
            next;
        }
        if (/${l}Thanatos$cl(?:[ _]+Destroyer)?:?\s*x?$nr/i) {
            $self->set_ships(Tha => $q);
            next;
        }
        if (/${l}Zeus$cl:?\s+x?$nr/i) {
            $self->set_ships(Zeu => $q);
            next;
        }

        if ($is_target) {
            if (/${l}(?:Hephaestus|Titan)$cl(?:[ _]Attack[ _]Platform)?:?\s*x?$nr/i) {
                $self->set_ships(Hep => $q);
                next;
            }
            if (/${l}Missile(?:[ _]Battery)?:?\s+x?$nr/i) {
                $self->set_ships(Mis => $q);
                next;
            }
            if (/${l}Laser(?:[ _]Cannon)?:?\s+x?$nr/i) {
                $self->set_ships(Las => $q);
                next;
            }
            if (/${l}(?:Space[ _])?Mine:?\s+x?$nr/i) {
                $self->set_ships(Min => $q);
                next;
            }
            if (/${l}Pulse(?:[ _]Cannon)?:?\s+x?$nr/i) {
                $self->set_ships(Pul => $q);
                next;
            }
            if (/${l}Particle(?:[ _]Cannon)?:?\s+x?$nr/i) {
                $self->set_ships(Par => $q);
                next;
            }
            if (/${l}Large(?:[ _]Decoy)?:?\s+x?$nr/i) {
                $self->set_ships(Lar => $q);
                next;
            }
            if (/${l}Decoy:?\s+x?$nr/i) {
                $self->set_ships(Dec => $q);
                next;
            }
            if (/${l}Gauss(?:[ _]Cannon)?:?\s+x?$nr/i) {
                $self->set_ships(Gau => $q);
                next;
            }
            if (/${l}Plasma(?:[ _]Cannon)?:?\s+x?$nr/i) {
                $self->set_ships(Pla => $q);
                next;
            }
        }

        if (/${l}Armor Tech:?\s+$nr/i) {
            $self->{armor} = $q;
            next;
        }
        if (/${l}Weapons Tech:?\s+$nr/i) {
            $self->{weapon} = $q;
            next;
        }
        if (/${l}Shield Tech:?\s+$nr/i) {
            $self->{shield} = $q;
            next;
        }
        if (/${l}Jet Drive:?\s+$nr/i) {
            $self->{jet} = $q;
            next;
        }
        if (/${l}Pulse Drive:?\s+$nr/i) {
            $self->{pulse} = $q;
            next;
        }
        if (/${l}Warp Drive:?\s+$nr/i) {
            $self->{warp} = $q;
            next;
        }

        if (/\[(\d+):(\d+):(\d+)([em]?)\]/) {
            $self->{location} = SFC::Planet->new($1, $2, $3, $4);
            next;
        }
    }

    close $f;
}

sub effect {
#    my $self = shift;
#    my ($eff, $ship) = @_;
#    $self->{$ship->type()}{effect} += $eff;
     $_[0]->{$_[2]->type}{effect} += $_[1];
}


sub id { $_[0]->{id} }

sub weapon_tech { $_[0]->{weapon} }
sub shield_tech { $_[0]->{shield} }
sub armor_tech { $_[0]->{armor} }

sub jet_tech { $_[0]->{jet} }
sub pulse_tech { $_[0]->{pulse} }
sub warp_tech { $_[0]->{warp} }

sub ore_lost { $_[0]->{ore_lost} }
sub crystal_lost { $_[0]->{cry_lost} }
sub hydrogen_lost { $_[0]->{gen_lost} }
sub debris { $_[0]->{debris} }
sub deb_ore { $_[0]->{deb_ore} }
sub deb_cry { $_[0]->{deb_cry} }

sub ore_plunder { $_[0]->{ore_plunder} }
sub crystal_plunder { $_[0]->{cry_plunder} }
sub hydrogen_plunder { $_[0]->{gen_plunder} }

sub location { $_[0]->{location} }

1;
