
package SFC::NelderMead;

use strict;
use warnings qw{ all };

our (@ISA, @EXPORT_OK);
BEGIN {
    require Exporter;
    @ISA = qw{Exporter};
    @EXPORT_OK = qw{
            optimize
            set_clamp
        };
}

use List::Util qw{ reduce };

my $alpha = 1;    # reflection coefficient
my $gamma = 2;    # expansion coefficient
my $rho   = .5;   # contraction coefficient
my $sigma = .5;   # shrink coefficient
my $quant = .02;  # quantize changes in components to this fraction

my $clamp;

sub set_clamp {
    $clamp = $_[0];
}

sub optimize {
    my ($f, @x) = @_;

    my $n = $#x;
    my $fault = 0;

    local $SIG{INT} = sub { print "SIGINT received\n"; $fault += 2 * $n; };

    while ($fault < 2 * $n) {

        # 1. Order according to the values at the vertices:
        @x = sort { $f->($a) <=> $f->($b) } @x;

        my $spread = $f->($x[$n]) - $f->($x[$n / 2]);
        printf "spread: %g, uncertainty: %g\n", $spread, $f->($x[$n])->error->[0];
        if ($spread < $f->($x[$n])->error->[0]) {
            print "Value spread less than error in mean.\n";
            last;
        }

        my $x_n1 = shift @x;    # worst point
        # 2. Calculate xo, the center of gravity of all points except xn + 1.
        my $x_o = (reduce { $a + $b } @x) / $n;

        # 3. Reflection
        #      Compute reflected point
        my $xrr = (($x_o + $alpha * ($x_o - $x_n1))->scale_round($quant))->clamp($clamp);
        ++$fault if $f->($xrr, 1);

        #      If the reflected point is better than the second worst,
        #      but not better than the best
        if ($f->($xrr) > $f->($x[0]) && $f->($xrr) <= $f->($x[$n - 1])) {
            #      then obtain a new simplex by replacing the worst point
            #      xn+1 with the reflected point xr, and go to step 1.
            push @x, $xrr;
            print "Reflection\n";
            next;
        }

        # 4. Expansion
        #      If the reflected point is the best point so far
        if ($f->($xrr) > $f->($x[$n - 1])) {
            #      then compute the expanded point
            my $xer = (($x_o + $gamma * ($x_o - $x_n1))->scale_round($quant))->clamp($clamp);
            ++$fault if $f->($xer, 1);

            #      If the expanded point is better than the reflected point
            if ($f->($xer) > $f->($xrr)) {
                #      then obtain a new simplex by replacing the worst point
                #      xn + 1 with the expanded point xe, and go to step 1.
                push @x, $xer;
                print "Expansion by xe\n";
                next;
            }

            #      Else obtain a new simplex by replacing the worst point
            #      xn + 1 with the reflected point xr, and go to step 1.
            push @x, $xrr;
            print "Expansion by xr\n";
            next;
        }

        #      Else (i.e. reflected point is not better than second worst)
        #      continue at step 5.

        # 5. Contraction
        #      Compute contracted point
        my $xcr = (($x_n1 + $rho * ($x_o - $x_n1))->scale_round($quant))->clamp($clamp);
        ++$fault if $f->($xcr, 1);

        #      If the contracted point is better than the worst point
        if ($f->($xcr) > $f->($x_n1)) {
            #      then obtain a new simplex by replacing the worst point
            #      xn+1 with the contracted point xc, and go to step 1.
            push @x, $xcr;
            print "Contraction\n";
            next;
        }

        #      Else go to step 6.

        # 6. Reduction

        #      For all but the best point, replace the point with
        unshift @x, $x_n1;
        my $x_1 = pop @x;
        for my $x_i (@x) {
            $x_i = (($x_1 + $sigma * ($x_i - $x_1))->scale_round($quant))->clamp($clamp);
            ++$fault if $f->($x_i, 1);
        }
        push @x, $x_1;
        print "Reduction around $x_1 (", $f->($x_1)->number, ")\n";
    }
        #      go to step 1.

    @x = sort { $f->($a) <=> $f->($b) } @x;

    return $x[$n];
}


1;
