
package SFC::Statistics;

use utf8;

use strict;
use warnings qw{ all };

use Statistics::Descriptive;
use Number::WithError;

use SFC::ShipStats qw{ desc_stats };

use POSIX qw{ ceil };

# our (@ISA, @EXPORT_OK);
# BEGIN {
#     require Exporter;
#     @ISA = qw{Exporter};
#     @EXPORT_OK = qw{
#             init_stats
#             record_data
#             print_stats
#         };
# }

sub value {
    my ($ore, $cry, $gen) = @_;
    # return 0.6 * $ore + $cry + 1.5 * $gen;
    return 0.6 * $ore + $cry + 3.0 * $gen;
}

sub new {
    my $this = shift;
    my $class = ref($this) || $this;
    my $self = ref($this) ? { %$this } : {
        };
    bless $self, $class;

    ($self->{attackers}, $self->{defenders}, $self->{ore_plunder},
        $self->{cry_plunder}, $self->{gen_plunder}, $self->{target},
        $self->{include_debris},
        ) = @_;

    $self->{win_index} = Statistics::Descriptive::Full->new();

    if ($self->{include_debris}) {
        $self->{debris} = Statistics::Descriptive::Full->new();
        $self->{deb_ore} = Statistics::Descriptive::Full->new();
        $self->{deb_cry} = Statistics::Descriptive::Full->new();
    }

    $self->{att_overkill_taken} = Statistics::Descriptive::Full->new();
    $self->{att_ore_lost} = Statistics::Descriptive::Full->new();
    $self->{att_cry_lost} = Statistics::Descriptive::Full->new();
    $self->{att_gen_lost} = Statistics::Descriptive::Full->new();
    $self->{att_cargo} = Statistics::Descriptive::Full->new();
    $self->{att_cost} = Statistics::Descriptive::Full->new();

    $self->{def_overkill_taken} = Statistics::Descriptive::Full->new();
    $self->{def_ore_lost} = Statistics::Descriptive::Full->new();
    $self->{def_cry_lost} = Statistics::Descriptive::Full->new();
    $self->{def_gen_lost} = Statistics::Descriptive::Full->new();


    for my $fs (@{$self->{attackers}}) {
        my $s = {};
        for my $t ($fs->types()) {
            for my $k (desc_stats) {
                $s->{$t}{$k} = Statistics::Descriptive::Full->new();
            }
        }
        push @{$self->{att_stats}}, $s;
    }

    for my $fs (@{$self->{defenders}}) {
        my $s = {};
        for my $t ($fs->types()) {
            for my $k (desc_stats) {
                $s->{$t}{$k} = Statistics::Descriptive::Full->new();
            }
        }
        push @{$self->{def_stats}}, $s;
    }

    if ($self->{target}) {
        $self->{att_hydro_cost} = 0;
        for my $fs (@{$self->{attackers}}) {
            my $source = $fs->location();
            if ($source) {
                my $dist = $self->{target}->distance($source);
                my $hcbd = $fs->hydro_cost_breakdown($dist, 1.0);
                for my $t (keys %$hcbd) {
                    $fs->{$t}{gen_loss} = $hcbd->{$t};
                    $fs->{$t}{loss} = 1.5 * $hcbd->{$t};
                    $self->{att_hydro_cost} += $hcbd->{$t};
                }
            }
            $self->{att_hydro_cost} = ceil($self->{att_hydro_cost});
        }
    }

    return $self;
}


sub record_data {
    my $self = shift;

    my ($rounds, $att_res, $def_res) = @_;

    $self->{win_index}->add_data($rounds);

    my ($aol, $acl, $dol, $dcl, $dhl, $deb, $car, $dbo, $dbc, )
        = (0, 0, 0, 0, 0, 0, 0, 0, 0);
    my $ahl = $self->{att_hydro_cost};
    
    # $att_overkill_taken->add_data($att->overkill_taken());
    # $def_overkill_taken->add_data($def->overkill_taken());

    for (my $i = 0; $i < @{$self->{attackers}}; ++$i) {
        $aol += $att_res->[$i]->ore_lost();
        $acl += $att_res->[$i]->crystal_lost();
        $ahl += $att_res->[$i]->hydrogen_lost();
        if ($self->{include_debris}) {
            $deb += $att_res->[$i]->debris();
            $dbo += $att_res->[$i]->deb_ore();
            $dbc += $att_res->[$i]->deb_cry();
        }
        for my $t ($self->{attackers}[$i]->types()) {
            for my $k (desc_stats) {
                $self->{att_stats}[$i]{$t}{$k}->add_data($att_res->[$i]{$t}{$k});
            }
        }
        $car += $att_res->[$i]->cargo();
    }
    $self->{att_ore_lost}->add_data($aol);
    $self->{att_cry_lost}->add_data($acl);
    $self->{att_gen_lost}->add_data($ahl);
    $self->{att_cargo}->add_data($car);
    $self->{att_cost}->add_data(value($aol, $acl, $ahl));

    for (my $i = 0; $i < @{$self->{defenders}}; ++$i) {
        $dol += $def_res->[$i]->ore_lost();
        $dcl += $def_res->[$i]->crystal_lost();
        $dhl += $def_res->[$i]->hydrogen_lost();
        if ($self->{include_debris}) {
            $deb += $def_res->[$i]->debris();
            $dbo += $def_res->[$i]->deb_ore();
            $dbc += $def_res->[$i]->deb_cry();
        }
        for my $t ($self->{defenders}[$i]->types()) {
            for my $k (desc_stats) {
                $self->{def_stats}[$i]{$t}{$k}->add_data($def_res->[$i]{$t}{$k});
            }
        }
    }
    $self->{def_ore_lost}->add_data($dol);
    $self->{def_cry_lost}->add_data($dcl);
    $self->{def_gen_lost}->add_data($dhl);

    if ($self->{include_debris}) {
        $self->{debris}->add_data($deb);
        $self->{deb_ore}->add_data($dbo);
        $self->{deb_cry}->add_data($dbc);
    }
}


sub print_stats {
    my $self = shift;

    my $tee;
    open($tee, '|-:utf8', 'tee', '-a', 'result.log')
        or $tee = \*STDOUT;
    print $tee "\nATTACKING FLEET(S)\n";
    for (my $i = 0; $i < @{$self->{attackers}}; ++$i) {
        my $fs = $self->{attackers}[$i];
        print $tee $i, ':';
        my $source = $fs->location();
        if ($self->{target} && $source) {
            my $dist = $self->{target}->distance($source);
            print $tee '  distance = ', $dist;
            # print $tee ", time = ", ;
            print $tee ', hydro cost (@100%) = ', $fs->hydro_cost($dist, 1.0);
        }
        print $tee "\n";
        for my $t ($fs->types()) {
            print $tee $t, ":  ";
            for my $k (desc_stats) {
                print $tee $k, "=";
                printf $tee "%.1f±%.1f", $self->{att_stats}[$i]{$t}{$k}->mean(), 
                    $self->{att_stats}[$i]{$t}{$k}->standard_deviation(), 
            }
            # continue {
            #     print $tee ", ";
            # }

            # How well did variation in the number of ships destroyed across
            # battles match the Poission distribution?  (1.0 = very well)
            if (0 && $self->{att_stats}[$i]{$t}{ships_lost}->mean() > 0) {
                print $tee ' ',
                    $self->{att_stats}[$i]{$t}{ships_lost}->standard_deviation() ** 2
                    / $self->{att_stats}[$i]{$t}{ships_lost}->mean();
            }

            # Ship-type effectiveness?
            if (0) {
                my $losses = $self->{att_stats}[$i]{$t}{loss};
                my $effects = $self->{att_stats}[$i]{$t}{effect};
                my $ovklts = $self->{att_stats}[$i]{$t}{overkill_taken};
                my $ovklgs = $self->{att_stats}[$i]{$t}{overkill_given};
                print $tee "rel_eff1=";
                if (my $loss = $losses->mean()) { 
                    my $eff = $effects->mean(); 
                    printf $tee "%.1f±%.1f", $eff / $loss,
                        sqrt(($loss * $effects->standard_deviation()) ** 2
                                + ($eff * $losses->standard_deviation()) ** 2)
                            / $loss ** 2;
                }
                else {
                    print $tee 'inf';
                }
                print $tee ", ";
                # print $tee "rel_eff2=";
                # if (my $loss = $losses->mean()) { 
                #     my $eff = $effects->mean() + $ovklts->mean(); 
                #     my $eff_unc = sqrt($effects->standard_deviation() ** 2 +
                #             $ovklts->standard_deviation() ** 2); 
                #     printf $tee "%.1f±%.1f", $eff / $loss,
                #         sqrt(($loss * $eff_unc) ** 2
                #                 + ($eff * $losses->standard_deviation()) ** 2)
                #             / $loss ** 2;
                # }
                # else {
                #     print $tee 'inf';
                # }
                print $tee "rel_eff3=";
                if (my $loss = $losses->mean()) { 
                    my $eff = $effects->mean() + $ovklts->mean() - $ovklgs->mean(); 
                    my $eff_unc = sqrt($effects->standard_deviation() ** 2 +
                            $ovklts->standard_deviation() ** 2 +
                            $ovklgs->standard_deviation() ** 2); 
                    printf $tee "%.1f±%.1f", $eff / $loss,
                        sqrt(($loss * $eff_unc) ** 2
                                + ($eff * $losses->standard_deviation()) ** 2)
                            / $loss ** 2;
                }
                else {
                    print $tee 'inf';
                }
            }

            print $tee "\n";
        }
    }

    print $tee "\nDEFENDING FLEET(S)\n";
    for (my $i = 0; $i < @{$self->{defenders}}; ++$i) {
        print $tee $i, ":\n";
        for my $t ($self->{defenders}[$i]->types()) {
            print $tee $t, ":  ";
            for my $k (desc_stats) {
                print $tee $k, "=";
                printf $tee "%.1f±%.1f", $self->{def_stats}[$i]{$t}{$k}->mean(), 
                    $self->{def_stats}[$i]{$t}{$k}->standard_deviation(), 
            }
            continue {
                print $tee ", ";
            }
            print $tee "\n";
        }
    }

    print $tee "Sims run: ", $self->{win_index}->count, "\n";

    printf $tee "Attack success index:  %d (%.2f±%.2f)\n",
            $self->{win_index}->median,
            $self->{win_index}->mean, $self->{win_index}->standard_deviation;

    my $plunder_value;
    my $debris_value;

    printf $tee "Attack cost:  %d±%d min=%d max=%d\n", $self->{att_cost}->mean,
            $self->{att_cost}->standard_deviation,
            $self->{att_cost}->min, $self->{att_cost}->max;

    printf $tee "Attacker available cargo space:  %d±%d  Minimum:  %d\n",
            $self->{att_cargo}->mean, $self->{att_cargo}->standard_deviation,
            $self->{att_cargo}->min;

    if ($self->success && defined $self->{ore_plunder}) {
        my $plunder_tot = $self->{ore_plunder} + $self->{cry_plunder} + $self->{gen_plunder};
        $plunder_value = value($self->{ore_plunder}, $self->{cry_plunder}, $self->{gen_plunder}); 
        printf $tee "Plunder available:  %d tot, value: %d\n",
                $plunder_tot, $plunder_value;
        
        my $plunder_diff = $plunder_tot - $self->{att_cargo}->min;
        if ($plunder_diff > 0) {
            print $tee 'May need ', $plunder_diff, " additional cargo space\n";
            print $tee 'E.g., ', ceil($plunder_diff / 5000), " Atlas, ",
            ceil($plunder_diff / 25000), " Hercs, or ",
            ceil($plunder_diff / 125000), " Carmanor.\n",
        }
    }

    printf $tee "Attacker losses:  %d±%d ore, %d±%d crystal, %d±%d hydrogen\n",
            $self->{att_ore_lost}->mean(),
            $self->{att_ore_lost}->standard_deviation(),
            $self->{att_cry_lost}->mean(),
            $self->{att_cry_lost}->standard_deviation(),
            $self->{att_gen_lost}->mean(),
            $self->{att_gen_lost}->standard_deviation();

    printf $tee "Defender losses:  %d±%d ore, %d±%d crystal, %d±%d hydrogen\n",
            $self->{def_ore_lost}->mean(),
            $self->{def_ore_lost}->standard_deviation(),
            $self->{def_cry_lost}->mean(),
            $self->{def_cry_lost}->standard_deviation(),
            $self->{def_gen_lost}->mean(),
            $self->{def_gen_lost}->standard_deviation();


    if ($self->{include_debris}) {
        printf $tee
        "Total debris:  %d±%d   Maximum debris:  %d (%d Dio | %d Zag)\n",
                $self->{debris}->mean(), $self->{debris}->standard_deviation(),
                $self->{debris}->max(), ceil($self->{debris}->max() / 20000),
                ceil($self->{debris}->max() / 5000);

        $debris_value = value($self->{deb_ore}->mean(),
                            $self->{deb_cry}->mean(), 0);
        printf $tee "Ore debris:  %d±%d   Crystal debris:  %d±%d  Value:  %d\n",
                $self->{deb_ore}->mean(),
                $self->{deb_ore}->standard_deviation(),
                $self->{deb_cry}->mean(),
                $self->{deb_cry}->standard_deviation(),
                $debris_value;
        print $tee "Profit estimate includes debris.\n";
    }

    print $tee 'Profit estimate: ', $self->profit(), "\n";
    close $tee;
}

sub profit {
    my $self = shift;

    my $att_cost_n = Number::WithError->new($self->{att_cost});
    my $att_carg_n = Number::WithError->new($self->{att_cargo});

    my $plunder_value = 0;
    if ($self->success && defined $self->{ore_plunder}) {
        $plunder_value = value(plunder_taken($att_carg_n, $self->{ore_plunder},
                $self->{cry_plunder}, $self->{gen_plunder})); 
    }

    my $att_ore_n = Number::WithError->new($self->{att_ore_lost});
    my $att_cry_n = Number::WithError->new($self->{att_cry_lost});
    my $att_gen_n = Number::WithError->new($self->{att_gen_lost});
    my $att_loss_val = value($att_ore_n, $att_cry_n, $att_gen_n);

    if ($self->{include_debris}) {
        my $deb_ore_n = Number::WithError->new($self->{deb_ore});
        my $deb_cry_n = Number::WithError->new($self->{deb_cry});
        my $debris_value = value($deb_ore_n, $deb_cry_n, 0);

        return $plunder_value + $debris_value - $att_loss_val;
    }

    return $plunder_value - $att_loss_val;
}


# Assumes res already been divided by 2 (Step 1).
sub plunder_taken {
    my ($space, $o, $c, $h) = @_;

    # Step 2:  If the available cargo space is higher than the amount of
    # plunderable resources on the planet, simply load them all and finish.
    return($o, $c, $h) if $space >= $o + $c + $h;

    # Step 3:  Attempt to load all 3 resources equally.
    # Step 3a: find the resource with the smallest amount.
    my @in = sort { $a->[1] <=> $b->[1] } ( [ 0, $o ], [ 1, $c ], [ 2, $h ] );

    # Step 3b: If that amount is greater than or equal to 1/3 of total cargo
    # capacity, then fill the cargo equally with each resource and finish.
    my $spcd = $space / 3;
    if ($in[0][1] >= $spcd) {
        return($spcd, $spcd, $spcd);
    }

    # Step 4:  the smallest resource is less than 1/3 of total cargo capacity;
    # load that amount of each resource into cargo.
    my @out = ($in[0][1], $in[0][1], $in[0][1]);
    $space -= 3 * $in[0][1];
    $in[1][1] -= $in[0][1];
    $in[2][1] -= $in[0][1];

    # Step 5:  Attempt to fill cargo space equally with the 2 remaining
    # resources.
    $spcd = $space / 2;
    if ($in[1][1] >= $spcd) {
        $out[$in[1][0]] += $spcd;
        $out[$in[2][0]] += $spcd;
        return @out;
    }

    # Step 5a:  If the smaller resource would take up less than half of the
    # remaining cargo space then fill the cargo space with all of that resource.
    $out[$in[1][0]] += $in[1][1];
    $space -= $in[1][1];

    # Step 5b: Then fill any remaining space with the third resource (the one
    # there was the most of.)
    $out[$in[2][0]] += $space;
    return @out;
}


sub success {
    my $self = shift;

    $self->{win_index}->mean > 3 * $self->{win_index}->standard_deviation;
}

sub break_even {
    my $self = shift;

    $self->{win_index}->mean > 3 * $self->{win_index}->standard_deviation;
}

sub success_index {
    my $self = shift;

    return Number::WithError->new($self->{win_index});
}

sub windex {
    my $self = shift;

    return $self->{win_index}->median;
}

1;
