#!/usr/bin/perl

use utf8;
# use locale;
binmode STDOUT, ':utf8';
binmode STDIN,  ':utf8';

use strict;
use warnings qw{ all };

use POSIX qw{ ceil };
use Statistics::Descriptive;
use Getopt::Long;

use SFC::FleetSummary;
use SFC::Simulate qw{ simulate };
use SFC::ShipStats qw{ desc_stats };



my $attackers = []; #  SFC::FleetSummary;
my $defenders = []; #  SFC::FleetSummary;

my $ore_plunder;
my $cry_plunder;
my $gen_plunder;
my $target;
my $att_hydro_cost;

#
# Get options
#

sub read_fsf {
    my ($name, $value) = @_;
    my $fs = new SFC::FleetSummary;
    $fs->read_file($value);

    if ($name eq 'att') {
        push @$attackers, $fs;
    }
    else {
        push @$defenders, $fs;
        $ore_plunder ||= $fs->ore_plunder();
        $cry_plunder ||= $fs->crystal_plunder();
        $gen_plunder ||= $fs->hydrogen_plunder();
        $target ||= $fs->location();
    }
}

my $sims = 100;

GetOptions( 'att=s@' => \&read_fsf, 
            'def=s@' => \&read_fsf,
            'num=i'  => \$sims,
    ) or die 'Error:  parsing options';



#
# Initialize statistics
#
my $win_index = Statistics::Descriptive::Full->new();
my $debris = Statistics::Descriptive::Full->new();

my $att_overkill_taken = Statistics::Descriptive::Full->new();
my $att_ore_lost = Statistics::Descriptive::Full->new();
my $att_cry_lost = Statistics::Descriptive::Full->new();
my $att_gen_lost = Statistics::Descriptive::Full->new();
my $att_cargo = Statistics::Descriptive::Full->new();
my $att_cost = Statistics::Descriptive::Full->new();

my $def_overkill_taken = Statistics::Descriptive::Full->new();
my $def_ore_lost = Statistics::Descriptive::Full->new();
my $def_cry_lost = Statistics::Descriptive::Full->new();
my $def_gen_lost = Statistics::Descriptive::Full->new();


my @att_stats;
for my $fs (@$attackers) {
    my $s = {};
    for my $t ($fs->types()) {
        for my $k (desc_stats) {
            $s->{$t}{$k} = Statistics::Descriptive::Full->new();
        }
    }
    push @att_stats, $s;
}

my @def_stats;
for my $fs (@$defenders) {
    my $s = {};
    for my $t ($fs->types()) {
        for my $k (desc_stats) {
            $s->{$t}{$k} = Statistics::Descriptive::Full->new();
        }
    }
    push @def_stats, $s;
}

if ($target) {
    for my $fs (@$attackers) {
        my $source = $fs->location();
        if ($source) {
            my $dist = $target->distance($source);
            my $hcbd = $fs->hydro_cost_breakdown($dist, 1.0);
            for my $t (keys %$hcbd) {
                $fs->{$t}{gen_loss} = $hcbd->{$t};
                $fs->{$t}{loss} = 1.5 * $hcbd->{$t};
                $att_hydro_cost += $hcbd->{$t};
            }
        }
        $att_hydro_cost = ceil($att_hydro_cost);
    }
}


#
# Main loop
#
for (my $i = $sims; $i > 0; --$i) {

    my ($rounds, $att_res, $def_res) = simulate($attackers, $defenders);

    $win_index->add_data($rounds);

    my ($aol, $acl, $dol, $dcl, $dhl, $deb, $car, )
        = (0, 0, 0, 0, 0, 0, 0);
    my $ahl = $att_hydro_cost;
    
    # $att_overkill_taken->add_data($att->overkill_taken());
    # $def_overkill_taken->add_data($def->overkill_taken());

    for (my $i = 0; $i < @$attackers; ++$i) {
        $aol += $att_res->[$i]->ore_lost();
        $acl += $att_res->[$i]->crystal_lost();
        $ahl += $att_res->[$i]->hydrogen_lost();
        $deb += $att_res->[$i]->debris();
        for my $t ($attackers->[$i]->types()) {
            for my $k (desc_stats) {
                $att_stats[$i]{$t}{$k}->add_data($att_res->[$i]{$t}{$k});
            }
        }
        $car += $att_res->[$i]->cargo();
    }
    $att_ore_lost->add_data($aol);
    $att_cry_lost->add_data($acl);
    $att_gen_lost->add_data($ahl);
    $att_cargo->add_data($car);
    $att_cost->add_data(.6 * $aol + $acl + 1.5 * $ahl);

    for (my $i = 0; $i < @$defenders; ++$i) {
        $dol += $def_res->[$i]->ore_lost();
        $dcl += $def_res->[$i]->crystal_lost();
        $dhl += $def_res->[$i]->hydrogen_lost();
        $deb += $def_res->[$i]->debris();
        for my $t ($defenders->[$i]->types()) {
            for my $k (desc_stats) {
                $def_stats[$i]{$t}{$k}->add_data($def_res->[$i]{$t}{$k});
            }
        }
    }
    $def_ore_lost->add_data($dol);
    $def_cry_lost->add_data($dcl);
    $def_gen_lost->add_data($dhl);

    $debris->add_data($deb);
}

#
# Print results
#
printf "Attack success index:  %.2f±%.2f\n", $win_index->mean,
        $win_index->standard_deviation;

#        print "Attacker:\n* ", join("\n* ", $attacker->summary_str()), "\n";
#        print "Defender:\n* ", join("\n* ", $defender->summary_str()), "\n";

printf "Attack cost:  %.2f±%.2f min=%d max=%d\n", $att_cost->mean,
        $att_cost->standard_deviation, $att_cost->min, $att_cost->max;

printf "Attacker available cargo space:  %.2f±%.2f  Minimum:  %d\n",
        $att_cargo->mean, $att_cargo->standard_deviation, $att_cargo->min;

if (defined $ore_plunder) {
    # printf "Plunder available:  %d ore, %d cry, %d gen, %d tot, value: %d\n",
    printf "Plunder available:  %d tot, value: %d\n",
            # $ore_plunder, $cry_plunder, $gen_plunder,
            $ore_plunder + $cry_plunder + $gen_plunder, 
            .6 * $ore_plunder + $cry_plunder + 1.5 * $gen_plunder; 
}

printf "Attacker losses:  %d±%d ore, %d±%d crystal, %d±%d hydrogen\n",
        $att_ore_lost->mean(), $att_ore_lost->standard_deviation(),
        $att_cry_lost->mean(), $att_cry_lost->standard_deviation(),
        $att_gen_lost->mean(), $att_gen_lost->standard_deviation();

printf "Defender losses:  %d±%d ore, %d±%d crystal, %d±%d hydrogen\n",
        $def_ore_lost->mean(), $def_ore_lost->standard_deviation(),
        $def_cry_lost->mean(), $def_cry_lost->standard_deviation(),
        $def_gen_lost->mean(), $def_gen_lost->standard_deviation();

printf "Total debris:  %d±%d   Maximum debris:  %d (%d Dios)\n",
        $debris->mean(), $debris->standard_deviation(),
        $debris->max(), ceil($debris->max() / 20000);

# printf "Attacker overkill taken:  %d±%d\n", $att_overkill_taken->mean(),
#         $att_overkill_taken->standard_deviation();
# printf "Defender overkill taken:  %d±%d\n", $def_overkill_taken->mean(),
#         $def_overkill_taken->standard_deviation();

print "\nATTACKING FLEET(S)\n";
for (my $i = 0; $i < @$attackers; ++$i) {
    my $fs = $attackers->[$i];
    print $i, ':';
    my $source = $fs->location();
    if ($target && $source) {
        my $dist = $target->distance($source);
        print '  distance = ', $dist;
        # print ", time = ", ;
        print ', hydro cost (@100%) = ', $fs->hydro_cost($dist, 1.0);
    }
    print "\n";
    for my $t ($fs->types()) {
        print $t, ":  ";
        for my $k (desc_stats) {
            print $k, "=";
            printf "%.1f±%.1f", $att_stats[$i]{$t}{$k}->mean(), 
                $att_stats[$i]{$t}{$k}->standard_deviation(), 
        }
        continue {
            print ", ";
        }

        my $losses = $att_stats[$i]{$t}{loss};
        my $effects = $att_stats[$i]{$t}{effect};
        my $ovklts = $att_stats[$i]{$t}{overkill_taken};
        my $ovklgs = $att_stats[$i]{$t}{overkill_given};
        print "rel_eff1=";
        if (my $loss = $losses->mean()) { 
            my $eff = $effects->mean(); 
            printf "%.1f±%.1f", $eff / $loss,
                sqrt(($loss * $effects->standard_deviation()) ** 2
                        + ($eff * $losses->standard_deviation()) ** 2)
                    / $loss ** 2;
        }
        else {
            print 'inf';
        }
        print ", ";
        # print "rel_eff2=";
        # if (my $loss = $losses->mean()) { 
        #     my $eff = $effects->mean() + $ovklts->mean(); 
        #     my $eff_unc = sqrt($effects->standard_deviation() ** 2 +
        #             $ovklts->standard_deviation() ** 2); 
        #     printf "%.1f±%.1f", $eff / $loss,
        #         sqrt(($loss * $eff_unc) ** 2
        #                 + ($eff * $losses->standard_deviation()) ** 2)
        #             / $loss ** 2;
        # }
        # else {
        #     print 'inf';
        # }
        print "rel_eff3=";
        if (my $loss = $losses->mean()) { 
            my $eff = $effects->mean() + $ovklts->mean() - $ovklgs->mean(); 
            my $eff_unc = sqrt($effects->standard_deviation() ** 2 +
                    $ovklts->standard_deviation() ** 2 +
                    $ovklgs->standard_deviation() ** 2); 
            printf "%.1f±%.1f", $eff / $loss,
                sqrt(($loss * $eff_unc) ** 2
                        + ($eff * $losses->standard_deviation()) ** 2)
                    / $loss ** 2;
        }
        else {
            print 'inf';
        }

        print "\n";
    }
}

print "\nDEFENDING FLEET(S)\n";
for (my $i = 0; $i < @$defenders; ++$i) {
    print $i, ":\n";
    for my $t ($defenders->[$i]->types()) {
        print $t, ":  ";
        for my $k (desc_stats) {
            print $k, "=";
            printf "%.1f±%.1f", $def_stats[$i]{$t}{$k}->mean(), 
                $def_stats[$i]{$t}{$k}->standard_deviation(), 
        }
        continue {
            print ", ";
        }
        print "\n";
    }
}
